/**************************************************
opengate server
 module for Controling ipfw 

Copyright (C) 1999 Yoshiaki Watanabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Email: watanaby@is.saga-u.ac.jp
**************************************************/


#include "opengatesrv.h"

char ruleNumber[WORDMAXLN];  /* ipfw rule number in string form */

int getRuleNumber(char *clientAddr);
int GetRuleNumber(char *clientAddr);
static void sigFunc(int signo);

/******************************************************************/
/* open gate for clientAddr (nnn.nnn.nnn.nnn)                     */
/******************************************************************/
int openClientGate(char *clientAddr, char *userid, char *macAddr, char *userProperty)
{
  int st;
  int fd;
  int ret;
  int retFork;
  int retNum;

  Sigfunc *defaultSigFunc;

  /* exclusive exec of ipfw to avoid overlapped rule number */

  /**** prepare ****/
  /* open lockfile */
  fd=Open(LOCKFILE, O_RDWR|O_CREAT, 
	  S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
  /* set timeout */
  defaultSigFunc=Signal(SIGALRM, sigFunc);
  alarm(LOCKTIMEOUT);
  /* lock */
  Lock(fd);
  /* reset timeout */
  Signal(SIGALRM, defaultSigFunc);
  alarm(0);

  /**** read rules ****/
  if((retNum=GetRuleNumber(clientAddr))<0){
    /* fail then unlock */
    Unlock(fd);
    Close(fd);
    if(retNum==-3) return 1; /* ipfw returns abnormal response */
    else return -1; /* already opened or rules are full */
  }

  /**** write rules ****/

  if(GetUseFwScript()==1){
    /********** use perl script to control firewall ************/

    /* fork */
    if((retFork=Fork())==0){
      /* child proc */
      if(execl(GetFwScriptPath(),GetFwScript(),IPFWPATH,ruleNumber,clientAddr,
	       userid,macAddr,userProperty,(char *)0)==-1){
	      err_quit("ERR in comm-ipfw: execl error no= %d", errno);
      }
      exit(1);
    }
    
    /* parent process */
    if(retFork==-1){
      err_msg("ERR in comm-ipfw: fork error no= %d", errno);
    }

    /* wait child end */
    wait(&st);
    if(WIFEXITED(st)){
      ret=WEXITSTATUS(st);
      if(ret) err_msg("ERR in comm-ipfw: child process returns error %d", ret);
    } else{
      ret=1;
      err_msg("ERR in comm-ipfw: child process does not exited by exit call");
    }

    Unlock(fd);   /* lock is not necessary in following exec */
    Close(fd);    /* because reserved number is used */

  }else{
    /********** direct control of firewall **********************/
    /********** add outgoing ipfw rule for the client *************/
    /* fork */
    if((retFork=Fork())==0){
      /* child proc */
      if(execl(IPFWPATH,"ipfw","-q","add",ruleNumber,"allow","ip",
	       "from",clientAddr,"to","any",(char *)0)==-1){
	      err_quit("ERR in comm-ipfw: execl error no= %d", errno);
      }
      exit(1);
    }

    /* parent process */
    if(retFork==-1){
      err_msg("ERR in comm-ipfw: fork error no= %d", errno);
    }

    /* wait child end */
    wait(&st);
    if(WIFEXITED(st)){
      ret=WEXITSTATUS(st);
      if(ret) err_msg("ERR in comm-ipfw: child process returns error %d", ret);
    } else{
      ret=1;
      err_msg("ERR in comm-ipfw: child process does not exited by exit call");
    }

    Unlock(fd);   /* lock is not necessary in following exec */
    Close(fd);    /* because reserved number is used */
    
    /********** add incoming ipfw rule for the client *************/
    /* fork */
    if((retFork=Fork())==0){
      /* child proc */
      if(execl(IPFWPATH,"ipfw","-q","add",ruleNumber,"allow","ip",
	       "from","any","to",clientAddr,(char *)0)==-1){
	      err_quit("ERR in comm-ipfw: execl error no= %d", errno);
      }
      exit(1);
    }

    /* parent process */
    if(retFork==-1){
      err_msg("ERR in comm-ipfw: fork error no= %d", errno);
    }

    /* wait child end */
    wait(&st);
    if(WIFEXITED(st)){
      ret=WEXITSTATUS(st);
      if(ret) err_msg("ERR in comm-ipfw: child process returns error %d", ret);
    } else{
      ret=1;
      err_msg("ERR in comm-ipfw: child process does not exited by exit call");
    }

  }
  return ret;
}

/******************************************************************/
/* close gate for clientAddr (nnn.nnn.nnn.nnn)                    */
/******************************************************************/
void closeClientGate(void)
{
  int st;

  /********** del ipfw rule for the client *************/

  /* fork */
  if(Fork()==0){
    /* child proc */
    if(execl(IPFWPATH,"ipfw","del",ruleNumber,(char *)0)==-1){
      err_quit("ERR in comm-ipfw: execl error no= %d", errno);
    }
    exit(1);
  }
  /* wait child end */
  wait(&st);

  return;
}




/**************************************/
/* get unused ipfw rule number        */
/* error if addr is already in rules */ 
/**************************************/
int getRuleNumber(char *clientAddr)
{
  int fd[2];
  int piperead_fd,pipewrite_fd;
  int stdin_fd=0,stdout_fd=1,stderr_fd=2;
  char buf[BUFFMAXLN];
  int st;
  int num,newNum,readinNum;
  int eofFound;
  int abnormalRes;
  char *p;
  int ipfwmin;
  int ipfwmax;
  int ipfwinterval;

  /* create pipe */
  if(pipe(fd)!=0){
    err_msg("ERR in comm-ipfw: Cannot create pipe for ipfw\n");
    return -1;
  }
  piperead_fd=fd[0];
  pipewrite_fd=fd[1];

  /* fork */
  if(Fork()==0){
    /* child proc */

    /* connect pipeout to stdout & strerr */
    Close(stdout_fd);
    Close(stderr_fd);
    dup(pipewrite_fd);
    dup(pipewrite_fd);

    /* close other IO */
    Close(stdin_fd);
    Close(piperead_fd);
    Close(pipewrite_fd);

    /* exec proc */
    if(execl(IPFWPATH,"ipfw","list",(char *)0) == -1){ 
      err_quit("ERR in comm-ipfw: execlp error no= %d", errno);
    }
    exit(1);
  }
  /* parent proc */
  
  /* close unuse IO */
  Close(pipewrite_fd);
  
  /* search unused rule number in the list read from pipe */
  /* check duplication of clientAddr to existing rules */
  newNum=-1;
  readinNum=0;
  eofFound=0;
  abnormalRes=0;

  /* get rule range from config */
  ipfwmin=GetIpfwMin();
  ipfwmax=GetIpfwMax();
  ipfwinterval=GetIpfwInterval();

  for(num=ipfwmin;num<=ipfwmax;num+=ipfwinterval){

    /* skip rules smaller than num */
    while(readinNum<num){
      if(Readln(piperead_fd, buf, BUFFMAXLN-1)==0){
	/* EOF */
	eofFound=1;
	break;
      }
      if( sscanf(buf, "%d", &readinNum) !=1 ){
	err_msg("ERR in comm-ipfw: abnormal ipfw response[ %s ]", buf);
	abnormalRes=1;
      }
    }
    
    /* if get abnormal response, error is returned */ 
    if(abnormalRes){
      newNum=-3;
      break;
    }

    /* if no rules larger or equal to num, then search end */
    if(eofFound){
      if(newNum<0) newNum=num;
      break;
    }

    /* if num is already used, then check dup and go to next search */
    if(readinNum==num){
      /* if clientAddr is found in rule, then err exit */
      if(((p=(char*)strstr(buf+1,clientAddr))!=NULL)
	&& isspace(*(p-1))
	&& !isalnum(*(p+strlen(clientAddr)))){
	err_msg("ERR in comm-ipfw: overlapped request from %s", clientAddr);
	newNum=-2;
	break;
      }
      continue;
    }

    /* num is not used */
    if(newNum<0)newNum=num;
    if(readinNum>ipfwmax)break;
  }
  
  /* close pipe */
  Close(piperead_fd);
    
  /* wait child end */
  wait(&st);

  if(newNum==-1){
    err_msg("ERR in comm-ipfw: cannot get unused ipfw number");
  }

  snprintf(ruleNumber, WORDMAXLN, "%d", newNum); /* to string */

  return newNum;
}

/*******************************/
/* get packet count from ipfw  */
/*******************************/
int getPacketCount(void)
{
  int fd[2];
  int piperead_fd,pipewrite_fd;
  int stdin_fd=0,stdout_fd=1;
  char buf[BUFFMAXLN];
  int st;
  int rule;
  int packets,packetsSum;

  /* create pipe */
  if(pipe(fd)!=0){
    err_msg("ERR in comm-ipfw: Cannot create pipe for ipfw\n");
    return -1;
  }
  piperead_fd=fd[0];
  pipewrite_fd=fd[1];

  /* fork */
  if(Fork()==0){
    /* child proc */

    /* connect pipeout to stdout */
    Close(stdout_fd);
    dup(pipewrite_fd);

    /* close other IO */
    Close(stdin_fd);
    Close(piperead_fd);
    Close(pipewrite_fd);

    /* exec proc */
    if(execl(IPFWPATH,"ipfw","-a","list",ruleNumber,(char *)0) == -1){ 
      err_quit("ERR in comm-ipfw: execl error no= %d", errno);
    }
  }
  /* parent proc */
  
  /* close unuse IO */
  Close(pipewrite_fd);
  
  /* search unused number in the list read from pipe */

  packetsSum=0;
    
  while(Readln(piperead_fd, buf, BUFFMAXLN-1)!=0){
    sscanf(buf, "%d %d", &rule, &packets);   /* get packet count */
    packetsSum+=packets;
  }

  /* close pipe */
  Close(piperead_fd);
    
  /* wait child end */
  wait(&st);

  return packetsSum;
}

/**********************************************/
/* function called by signal int              */
/**********************************************/
static void sigFunc(int signo)
{
  return;
}

/**********************************************/
/**********************************************/
int GetRuleNumber(char *clientAddr)
{
  int ret;

  if(DEBUG) err_msg("DEBUG:=>getRuleNumber(%s)",clientAddr);
  ret=getRuleNumber(clientAddr);
  if(DEBUG) err_msg("DEBUG:(%d)<=getRuleNumber( )",ret);

  return ret;
}

int OpenClientGate(char *clientAddr, char *userid, char *macAddr, char *userProperty)
{
  int ret;

  if(DEBUG) err_msg("DEBUG:=>openClientGate(%s,%s,%s,%s)",clientAddr,userid,macAddr,userProperty);
  ret=openClientGate(clientAddr, userid, macAddr, userProperty);
  if(DEBUG) err_msg("DEBUG:(%d)<=openClientGate( )",ret);

  return ret;
}

void CloseClientGate(void)
{
  if(DEBUG) err_msg("DEBUG:=>closeClientGate( )");
  closeClientGate();
  if(DEBUG) err_msg("DEBUG:<=closeClientGate( )");
}


int GetPacketCount(void)
{
  int ret;

  if(DEBUG) err_msg("DEBUG:=>getPacketCount( )");
  ret=getPacketCount();
  if(DEBUG) err_msg("DEBUG:(%d)<=getPacketCount( )",ret);

  return ret;
}

