/**************************************************
opengate server main

Copyright (C) 1999 Yoshiaki Watanabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Email: watanaby@is.saga-u.ac.jp
**************************************************/

#include	"opengatesrv.h"

void putCloseMsg(time_t timeOut, time_t timeIn);
void PutCloseMsg(time_t timeOut, time_t timeIn);
void closeExit(int signo);

extern char ruleNumber[WORDMAXLN];  /* ipfw rule number in string form */

char clientAddr[ADDRMAXLN]="";  /* client addr (nnn.nnn.nnn.nnn) */
char macAddr[ADDRMAXLN]="?";    /* client MAC address */

char userid[USERMAXLN];
char useridshort[USERMAXLN];/* userID before @ mark(cut off serverID) */
char serverid[USERMAXLN];
char userProperty[BUFFMAXLN];
time_t timeIn, timeOut;

/***************************************************/
/*  main routine called as cgi from Web server     */
/***************************************************/
int  main(int argc, char **argv)
{
  char password[PASSMAXLN];
  int port;
  int dummyfd[2];
  int pid;
  int ret;
  int duration;     /* requested usage duration */
 
  /* start log */
  errToSyslog(ERRORTOSYSLOG);     /* output of err_xxx() to syslogd */
  openlog(argv[0], LOG_PID, FACILITY);

  /* get POST data */
  GetPostData(userid, password, &duration);

  /* split user@server to user and server */
  SplitId(userid, useridshort, serverid);

  /* read config file and setup parameters buffer */
  CreateDbBuffer(serverid);

  /* get information of client machine */
  GetClientAddr(clientAddr);
  if(GetEnableMacAddr()==1) GetMacAddr(clientAddr, macAddr);

  /* authenticate the user with authentication server */
  if(!AuthenticateUser(useridshort, password)){
    PutClientDeny();
    err_msg("DENY: user %s from %s at %s", userid, clientAddr, macAddr);
    return 0;
  }

  /* clear password */
  bzero(password, PASSMAXLN);

  /* get user property from user database (if you edit comm-userdb.c) */
  if(!GetUserProperty(userid, userProperty)){
    PutClientMsg("Error: You are denied.");
    err_msg("DENY: user %s from %s at %s (ill-property)",
	    userid, clientAddr, macAddr);
    return 0;
  }

  /* entered duration is restricted to permitted range */
  if(duration <= 0){
    duration=GetDurationDefault();
  }else if(duration > GetDurationMax()){
    duration=GetDurationMax();
  }

  /* open firewall for the client */
  if((ret=OpenClientGate(clientAddr, userid, macAddr, userProperty))<0){
    PutClientMsg("Error: Please End Web and Retry");
    return 0;
  }else if(ret>0){
    PutClientMsg("Error: Please contact to the administrator");
    return 0;
  }

  timeIn=time(NULL);
  err_msg("OPEN: user %s from %s at %s", userid, clientAddr, macAddr);

  /* set (ruleNumber,userid,clientAddr) in process title */
  setproctitle("%s,%s,%s", ruleNumber, userid, clientAddr);

  /* set terminate signal handler */
  Signal(SIGTERM, closeExit);

  /* get temporary port for server-listen */
  port=GetListenPort();
  if(port<0){
    err_msg("ERR in main: cannot get unused listen port");
    PutClientMsg("Error: Please Call Administrator");
    closeExit(1);
  }

  if((pid=Fork())!=0){
    /** parent process **/
    /* send java */
    PutJavaApplet(userid, port, pid);

    /* detach from Web server */
    return 0;
  }
  /** child process **/

  /* detach from Web server */
  Close(0);Close(1);  /* detach stdin and out pipe connected to Web */
  Close(2);           /* detach stderr */
  Pipe(dummyfd);      /* connect dummy pipe for stdin and out */

  /* connect to the java applet */
  if(WaitAppletConnect(userid, clientAddr, duration, macAddr)==1){

    /* wait until the user quit */
    WaitClientClose();
  }
  /* close firewall and exit */
  closeExit(1);

  return 0;
}


/* calc connection duration and put it out */

void putCloseMsg(time_t timeOut, time_t timeIn)
{
  double time;
  int hour,min,sec;

  time=difftime(timeOut,timeIn);
  hour=time/60/60;
  min=(time-hour*60*60)/60;
  sec=(time-hour*60*60-min*60);
  err_msg("CLOS: user %s from %s at %s ( %02d:%02d:%02d )",
	  userid, clientAddr, macAddr, hour,min,sec);
  return;
}

/*****************************/
/* At termination, call this */
/*****************************/
void closeExit(int signo)
{
  Signal(SIGTERM, SIG_DFL);
  QuitClient();
  CloseClientGate();
  timeOut=time(NULL);
  PutCloseMsg(timeOut,timeIn);
  if(DEBUG) err_msg("DEBUG:terminated");
  exit(1);  
}

/*****************************/
/*****************************/
void PutCloseMsg(time_t timeOut, time_t timeIn)
{
  if(DEBUG) err_msg("DEBUG:=>putCloseMsg( )");
  putCloseMsg(timeOut,timeIn);
  if(DEBUG) err_msg("DEBUG:<=putCloseMsg( )");
}

